@@ -1,32 +1,28 @@
-isopam <-
-  function (dat, c.fix = FALSE, c.opt = TRUE, c.max = 6, 
-            l.max = FALSE, stopat = c(1,7), sieve = TRUE, 
-            Gs = 3.5, ind = NULL, centers = NULL, distance = 'bray', 
-            k.max = 100, d.max = 7, ..., juice = FALSE) 
+isopam <- function (dat, c.fix = FALSE, c.opt = TRUE, c.max = 6,
+                    l.max = FALSE, stopat = c(1,7), sieve = TRUE,
+                    Gs = 3.5, ind = NULL, centers = NULL, distance = 'bray',
+                    k.max = 100, d.max = 7, ..., juice = FALSE)
 {
 
-  if (distance != "bray" & distance != "jaccard") 
-        require(proxy) || stop("needs package proxy")
-  
   if (!is.null (centers)) c.fix <- length (centers)
-  
+
   ## Make backwards compatible (this is mainly for use with Juice)
-  if (!is.null (list (...)$fixed.number)) c.fix <- list (...)$fixed.number 
-  if (!is.null (list (...)$c.num)) c.fix <- list (...)$c.num ## old Juice vers.!  
-  if (!is.null (list (...)$opt.number)) c.opt <- list (...)$opt.number   
-  if (!is.null (list (...)$max.number)) c.max <- list (...)$max.number 
-  if (!is.null (list (...)$max.level)) l.max <- list (...)$max.level 
-  if (!is.null (list (...)$thresh)) Gs <- list (...)$thresh  
-  if (!is.null (list (...)$filtered)) sieve <- list (...)$filtered 
-  ## Note that Gs is set to a default of 30 in old Juice versions. The setting 
+  if (!is.null (list (...)$fixed.number)) c.fix <- list (...)$fixed.number
+  if (!is.null (list (...)$c.num)) c.fix <- list (...)$c.num ## old Juice vers.!
+  if (!is.null (list (...)$opt.number)) c.opt <- list (...)$opt.number
+  if (!is.null (list (...)$max.number)) c.max <- list (...)$max.number
+  if (!is.null (list (...)$max.level)) l.max <- list (...)$max.level
+  if (!is.null (list (...)$thresh)) Gs <- list (...)$thresh
+  if (!is.null (list (...)$filtered)) sieve <- list (...)$filtered
+  ## Note that Gs is set to a default of 30 in old Juice versions. The setting
   ## is ignored here
-  
-  ## Prepare Juice session if applicable 
+
+  ## Prepare Juice session if applicable
   if (juice == TRUE) dir.create ('isopam', showWarnings = FALSE)
 
   ## Add fake 'sample names' if necessary
   if (is.null (rownames (dat)) == TRUE) rownames (dat) <- c(1:nrow(dat))
-   
+
   ## Add fake 'taxon names' if necessary
   if (is.null (colnames (dat)) == TRUE) colnames (dat) <- c(1:ncol(dat))
 
@@ -41,17 +37,17 @@ isopam <-
 
   ## In case of predefined indicators check their validity
   if (!is.null (ind))
-    if (sum (ind %in% colnames (dat)) > 0) sieve <- 'ind'      
-    else stop ('Predefined indicators not found')
-    
+    if (sum (ind %in% colnames (dat)) > 0) sieve <- 'ind'
+  else stop ('Predefined indicators not found')
+
   ## Initiate count
   count <- 1
-  
+
   ## Default minimum cluster number
   c.min <- 2
-  
+
   if (is.numeric (c.fix))
-  {  
+  {
     c.opt <- FALSE
     l.max <- 1
     if (c.fix < 2) stop ('c.fix < 2')
@@ -62,7 +58,7 @@ isopam <-
 
   ## ----------- core function ---------------------------------------------- ##
 
-  core <- function (xdat) 
+  core <- function (xdat)
   {
     IO.xdat <- xdat
     IO.xdat [IO.xdat > 0] <- 1
@@ -71,29 +67,28 @@ isopam <-
     N.xdat <- nrow (xdat)                         ## Total number of plots
     SP.xdat <- ncol (xdat)                        ## Total number of species
     frq.xdat <- t (as.matrix (colSums (IO.xdat))) ## Species frequencies
-    
+
     ## For William's correction
-    w3 <- N.xdat * ((1 / frq.xdat) + (1 / (N.xdat - frq.xdat))) - 1 
+    w3 <- N.xdat * ((1 / frq.xdat) + (1 / (N.xdat - frq.xdat))) - 1
 
     ## In case of predefined indicators: which columns?
     if (sieve == 'ind') xind <- which (colnames (xdat) %in% ind)
 
-    ## Distance matrix    
-    dst.xdat <- try (vegdist (xdat, method = distance), silent = TRUE)
+    ## Distance matrix
+    dst.xdat <- try (vegan::vegdist (xdat, method = distance), silent = TRUE)
     ## if vegan does not work try with package proxy
-    if (class (dst.xdat) == 'try-error')
-    { 
-      require (proxy) || stop ('needs package proxy')     
-      dst.xdat <- dist (xdat, method = distance)
-    }   
-    
+    if (methods::is(dst.xdat, 'try-error'))
+    {
+      dst.xdat <- proxy::dist (xdat, method = distance)
+    }
+
     ## Exit with dignity if N.xdat < 3
-    if (N.xdat < 3)                                                      
+    if (N.xdat < 3)
     {
       mess1 <- 'Not enough data.'
-      if (juice == TRUE) 
+      if (juice == TRUE)
       {
-      write.table (mess1, 'isopam/alert.txt', row.names = FALSE, 
+      write.table (mess1, 'isopam/alert.txt', row.names = FALSE,
         col.names = FALSE, quote = FALSE)
       }
       stop (mess1)
@@ -113,11 +108,11 @@ isopam <-
       ## Check min. k
       dm <- dmtr
 
-      is.na (dm) <- apply (dm, 2, function (xx) xx > 
+      is.na (dm) <- apply (dm, 2, function (xx) xx >
         xx [order (xx, na.last = TRUE) [a]])
 
       dm <- pmax (as.dist(dm), as.dist(t(dm)), na.rm = TRUE)
-      fragm <- vegan::distconnected (dm, toolong=0,  trace=FALSE)
+      fragm <- vegan::distconnected (dm, toolong = 0,  trace = FALSE)
       if (length (unique (fragm)) > 1)
         k.min <- k.min + 1
       else
@@ -126,63 +121,63 @@ isopam <-
 
     ## Adjust d.max if necessary
     if (d.max > N.xdat - 1) d.max <- N.xdat - 1
-    
+
     ## Adjust c.max if necessary
-  
-    if (c.opt == TRUE)  
+
+    if (c.opt == TRUE)
     {
       if (c.max > N.xdat - 1) c.max <- N.xdat - 1
       if (c.max < 2) stop ('c.max < 2')
-    }    
-    if (c.opt == FALSE && !is.numeric (c.fix)) c.max <- 2    
-            
+    }
+    if (c.opt == FALSE && !is.numeric (c.fix)) c.max <- 2
+
     ## Prepare progress bar
     ## Problem: Overestimation of d.loops in case of neg. eigenvalues in isomap
     b.loops <- (k.max - k.min) + 1
-    d.loops <- d.max - 1 
+    d.loops <- d.max - 1
     pb.mx <- (b.loops * d.loops)
     pb <- txtProgressBar (min = 0, max = pb.mx, char = '.',
       width = 45, style = 3)
-    
+
     ## Prepare output array
     out.array <- array (NA, dim = c(10, k.max, c.max))
 
     ## Need a place to store the calculations done in the g2 loop
     ## so they may be reused rather than recalculated.
     ## The +1 is due to zeros needing to be indexed as well.
-    
-    try (DDD_lookup_table <- 
-      array (NA_real_, c(N.xdat, N.xdat+1, N.xdat+1)), silent = TRUE)    
+
+    try (DDD_lookup_table <-
+      array (NA_real_, c(N.xdat, N.xdat + 1, N.xdat + 1)), silent = TRUE)
     underdrive <- FALSE
-    if (!exists ('DDD_lookup_table')) 
+    if (!exists ('DDD_lookup_table'))
     {
-      print ('Memory issues - shifting to low gear', quote = FALSE)    
+      print ('Memory issues - shifting to low gear', quote = FALSE)
       underdrive <- TRUE
     }
-      
+
     ## Using a matrix of logical values is a _hair_ faster inside the loop
     IO.xdat.logical <- IO.xdat != 0
-    
-    ## ----------- Begin with the big loops .... ---------------------------- ##
+
+    ## ----------- Start the big loops .... ---------------------------- ##
 
     for (b in c(k.min:k.max))  ## b-loop: Isomap k
-    {   
+    {
 
-      suppressWarnings (isom <- isomap (dst.xdat, ndim = d.max, k = b)) ## Isomap   
+      suppressWarnings (isom <- isomap (dst.xdat, ndim = d.max, k = b)) ## Isomap
       ## this is for post R-2.13 versions:
-      d.max.new <- min (sum (isom$eig > 0), ncol (isom$points))      
-      
-      for (d in 2:d.max.new)  
+      d.max.new <- min (sum (isom$eig > 0), ncol (isom$points))
+
+      for (d in 2:d.max.new)
       {
-            
+
         isodiss <- suppressWarnings (daisy (isom$points[,1:d], metric =
           'euclidean', stand = TRUE))
-        
+
         for (e in c.min:c.max) ## e-loop: Cluster no.
-        {                      
+        {
           ## --------- Partitioning (PAM) ------------------------------------ #
-          
-          if (!is.null (centers)) cl.iso <- pam (isodiss, k = e, medoids = centers, 
+
+          if (!is.null (centers)) cl.iso <- pam (isodiss, k = e, medoids = centers,
             diss = TRUE, do.swap = FALSE)
           else cl.iso <- pam (isodiss, k = e, diss = TRUE) ## PAM
           cl <- cl.iso$clustering                     ## Group affiliation
@@ -193,32 +188,32 @@ isopam <-
           if (underdrive == FALSE)
           {
             ## For Williams' correction
-            w1 <- N.xdat * sum (1 / ci) - 1 
+            w1 <- N.xdat * sum (1 / ci) - 1
             w2 <- 6 * N.xdat * (e - 1)
-            
+
             ## Compute G-values for species (code adapted from Lubomir Tichy)
-            
+
             gt <- matrix (NA, SP.xdat, 1) ## Matrix for G-test results
-  
+
             for (g1 in 1:SP.xdat)         ## g1-loop through species
-            {                    
+            {
               DDD <- 0
               spec_frq <- frq.xdat [g1]
-  
+
               ## For Williams' correction
               willi <- 1 + ((w1 * w3 [g1]) / w2)
-              
+
               ## Mask out entries in cl using appropriate col in IO.xdat
               groupids <- cl[IO.xdat.logical[,g1]]
-               
+
               for (g.fast in 1:e)             ## g.fast-loop through clusters
-              {                  
+              {
                 fra1 <- sum (groupids == g.fast)   ## Species occ. in cluster
                 Nj <- ci [g.fast]                  ## Cluster size
-  
+
                 ## Have we calculated this before?
                 DDDadd <- DDD_lookup_table [spec_frq, Nj+1, fra1+1]
-  
+
                 if (!is.na(DDDadd))
                 {
                 ## Already existed in the lookup table; use it.
@@ -227,16 +222,16 @@ isopam <-
                 else
                 {
                   ## so need to calculate it ...
-                  bom <- spec_frq / N.xdat 
-                  bim <- 1 - bom           
-                  fra0 <- Nj - fra1        
-                  bum <- fra1 / (Nj * bom)  
-                  bam <- fra0 / (Nj * bim) 
+                  bom <- spec_frq / N.xdat
+                  bim <- 1 - bom
+                  fra0 <- Nj - fra1
+                  bum <- fra1 / (Nj * bom)
+                  bam <- fra0 / (Nj * bim)
                   DDDadd <- 0
-                  if (!is.na (bum)) 
-                    if (bum > 0) DDDadd <- DDDadd + (fra1 * log (bum)) 
-                  if (!is.na (bam)) 
-                    if (bam > 0) DDDadd <- DDDadd + (fra0 * log (bam))             
+                  if (!is.na (bum))
+                    if (bum > 0) DDDadd <- DDDadd + (fra1 * log (bum))
+                  if (!is.na (bam))
+                    if (bam > 0) DDDadd <- DDDadd + (fra0 * log (bam))
                   DDD <- DDD + DDDadd
                   ## ... and store for next time
                   DDD_lookup_table [spec_frq, Nj + 1, fra1 + 1] <- DDDadd
@@ -245,30 +240,30 @@ isopam <-
               DDD <- DDD * 2
               gt [g1,] <- DDD / willi ## Williams' correction
             }
-  
+
             ## Standardization (Botta-Dukat et al. 2005)
             gt.ex <- e - 1                 ## Expected G
             gt.sd <- sqrt (2 * gt.ex)      ## Expected sd
             G <- (gt - gt.ex) / gt.sd
-            
+
             ## Using predefined indicators
             if (sieve == 'ind')
             {
-              glgth <- length (G [xind]) 
+              glgth <- length (G [xind])
               if (glgth == 0) out.array [d,b,e] <- NA
               else out.array [d,b,e] <- mean (G [xind])
             }
-            
+
             ## Averaging
             if (sieve == FALSE) out.array [d,b,e] <- mean (G)
-            
+
             ## Standard: Filtering and averaging
-            if (sieve == TRUE) 
+            if (sieve == TRUE)
             {
               ## Filtering by G
-              glgth <- length (G [G >= Gs]) 
+              glgth <- length (G [G >= Gs])
               if (glgth == 0) out.array [d,b,e] <- NA
-  
+
               ## Averaging
               else out.array [d,b,e] <- mean (G [G >= Gs]) * glgth
             }
@@ -279,93 +274,93 @@ isopam <-
           if (underdrive == TRUE)
           {
             ## For Williams' correction
-            w1 <- N.xdat * sum (1 / ci) - 1 
+            w1 <- N.xdat * sum (1 / ci) - 1
             w2 <- 6 * N.xdat * (e - 1)
-            
+
             ## Compute G-values for species (code adapted from Lubomir Tichy)
-            
+
             gt <- matrix (NA, SP.xdat, 1) ## Matrix for G-test results
-  
+
             for (g1 in 1:SP.xdat)         ## g1-loop through species
-            {                    
+            {
               DDD <- 0
               spec_frq <- frq.xdat [g1]
-  
+
               ## For Williams' correction
               willi <- 1 + ((w1 * w3 [g1]) / w2)
-              
+
               ## Mask out entries in cl using appropriate col in IO.xdat
-              groupids <- cl[IO.xdat.logical[,g1]]; 
-  
+              groupids <- cl[IO.xdat.logical[,g1]];
+
               for (g.slow in 1:e)             ## g.slow-loop through clusters
-              {                  
+              {
                 fra1 <- sum (groupids == g.slow)   ## Species occ. in cluster
                 Nj <- ci [g.slow]                  ## Cluster size
-                bom <- spec_frq / N.xdat 
-                bim <- 1 - bom           
-                fra0 <- Nj - fra1        
-                bum <- fra1 / (Nj * bom)  
-                bam <- fra0 / (Nj * bim) 
+                bom <- spec_frq / N.xdat
+                bim <- 1 - bom
+                fra0 <- Nj - fra1
+                bum <- fra1 / (Nj * bom)
+                bam <- fra0 / (Nj * bim)
                 DDDadd <- 0
-                if (!is.na (bum)) 
-                  if (bum > 0) DDDadd <- DDDadd + (fra1 * log (bum)) 
-                if (!is.na (bam)) 
-                  if (bam > 0) DDDadd <- DDDadd + (fra0 * log (bam))             
+                if (!is.na (bum))
+                  if (bum > 0) DDDadd <- DDDadd + (fra1 * log (bum))
+                if (!is.na (bam))
+                  if (bam > 0) DDDadd <- DDDadd + (fra0 * log (bam))
                 DDD <- DDD + DDDadd
               }
               DDD <- DDD * 2
               gt [g1,] <- DDD / willi ## Williams' correction
             }
-  
+
             ## Standardization (Botta-Dukat et al. 2005)
             gt.ex <- e - 1                 ## Expected G
             gt.sd <- sqrt (2 * gt.ex)      ## Expected sd
             G <- (gt - gt.ex) / gt.sd
-            
+
             ## Using predefined indicators
             if (sieve == 'ind')
             {
-              glgth <- length (G [xind]) 
+              glgth <- length (G [xind])
               if (glgth == 0) out.array [d,b,e] <- NA
               else out.array [d,b,e] <- mean (G [xind])
             }
-            
+
             ## Averaging
             if (sieve == FALSE) out.array [d,b,e] <- mean (G)
-            
+
             ## Standard: Filtering and averaging
-            if (sieve == TRUE) 
+            if (sieve == TRUE)
             {
               ## Filtering by G
-              glgth <- length (G [G >= Gs]) 
+              glgth <- length (G [G >= Gs])
               if (glgth == 0) out.array [d,b,e] <- NA
-  
+
               ## Averaging
               else out.array [d,b,e] <- mean (G [G >= Gs]) * glgth
             }
           }
-          
+
           ######################################################################
 
         }
         setTxtProgressBar (pb, ((b - k.min) * (d.max - 1)) + d - 1)
       }
     }
-    
+
     close (pb) ## Close progress bar
 
     ## ----------- End parameter search ------------------------------------- ##
-    
+
     solution <- TRUE
     out.array [is.na (out.array)] <- 0
-    
-    if (length (out.array [out.array > 0]) == 0) 
-    { 
+
+    if (length (out.array [out.array > 0]) == 0)
+    {
       mess2 <- 'No solution found with current settings'
-      
-      if (count == 1) 
+
+      if (count == 1)
       {
-        if (juice == TRUE) 
+        if (juice == TRUE)
         {
           write.table (mess2, 'isopam/alert.txt',
             row.names = FALSE, col.names = FALSE, quote = FALSE)
@@ -373,9 +368,9 @@ isopam <-
         stop ('No solution found with current settings')
       }
       else solution <- FALSE
-    }  
-    
-    if (solution == TRUE) 
+    }
+
+    if (solution == TRUE)
     {
       mn.iso <- max (out.array)
 
@@ -383,8 +378,8 @@ isopam <-
 
       wmx.iso <- which (out.array == mn.iso, arr.ind = TRUE) ## Cases with max
       colnames (wmx.iso) <- c('iso.dim', 'iso.k', 'clusters')
-      
-      ## In case of multiple best solutions select the one with max clusters, 
+
+      ## In case of multiple best solutions select the one with max clusters,
       ## max dims and max k (in this rank order)
 
       ## Select cases with maximum number of clusters
@@ -398,14 +393,14 @@ isopam <-
         (wmx.iso [,2])),], silent = TRUE)
 
       mc <- wmx.iso [3]; md <- wmx.iso [1]; mk <- wmx.iso [2]
-      
+
       ## ----------- Final run ---------------------------------------------- ##
-      suppressWarnings (isom <- isomap (dst.xdat, ndim = d.max, k = mk))                
+      suppressWarnings (isom <- isomap (dst.xdat, ndim = d.max, k = mk))
       d.iso <- daisy (isom$points[,1:md], metric = 'euclidean', stand = TRUE)
-      if (!is.null (centers)) cl.iso <- pam (d.iso, k = mc, medoids = centers, 
+      if (!is.null (centers)) cl.iso <- pam (d.iso, k = mc, medoids = centers,
         diss = TRUE, do.swap=FALSE)
       else cl.iso <- pam (d.iso, k = mc, diss = TRUE)
-      
+
       CLS <- cl.iso$clustering                 ## Group affiliation
       MDS <- cl.iso$medoids                    ## Medoids
       CLI <- t (cl.iso$clusinfo [,1])          ## Cluster size
@@ -413,38 +408,38 @@ isopam <-
       ############# method == G ################################################
 
       ## Contingency table
-      tab <- t (aggregate (IO.xdat, by = list (CLS), FUN = sum)[,-1]) 
+      tab <- t (aggregate (IO.xdat, by = list (CLS), FUN = sum)[,-1])
       ## Frequency table
       inf <- matrix (rep (CLI, SP.xdat), nrow = SP.xdat, byrow = TRUE)
-      FRQ <- tab / inf 
-  
+      FRQ <- tab / inf
+
       ## Prepare Williams' correction
-      w1 <- N.xdat * sum (1 / CLI) - 1 
+      w1 <- N.xdat * sum (1 / CLI) - 1
       w2 <- 6 * N.xdat * (mc - 1)
 
       ## Matrix for results
-      g.1 <- matrix (NA, SP.xdat, 1)  
+      g.1 <- matrix (NA, SP.xdat, 1)
 
-      for (g3 in 1:SP.xdat)           
-      {                             
+      for (g3 in 1:SP.xdat)
+      {
         willi <- 1 + ((w1 * w3 [g3]) / w2)
         DDD <- 0
-        bom <- frq.xdat [g3] / N.xdat 
-        bim <- 1 - bom                
-        rip <- tab [g3,]              
-        
-        for (g4 in 1:mc)              
-        {                           
-          fra1 <- rip [g4]            
-          Nj <- CLI [g4]              
-          fra0 <- Nj - fra1           
-          bum <- fra1 / (Nj * bom)              
-          bam <- fra0 / (Nj * bim)            
-  
+        bom <- frq.xdat [g3] / N.xdat
+        bim <- 1 - bom
+        rip <- tab [g3,]
+
+        for (g4 in 1:mc)
+        {
+          fra1 <- rip [g4]
+          Nj <- CLI [g4]
+          fra0 <- Nj - fra1
+          bum <- fra1 / (Nj * bom)
+          bam <- fra0 / (Nj * bim)
+
           ## adding up values
-          if (!is.na (bum))   
+          if (!is.na (bum))
             if (bum > 0) DDD <- DDD + (fra1 * log (bum))
-          if (!is.na (bam))   
+          if (!is.na (bam))
             if (bam > 0) DDD <- DDD + (fra0 * log (bam))
         }
         DDD <- DDD * 2
@@ -455,8 +450,8 @@ isopam <-
       gt.ex <- mc - 1                        ## Expected G
       gt.sd <- sqrt (2 * (mc - 1))           ## Expected sd
       sG <- (g.1 - gt.ex) / gt.sd
-      
-      ## Some analytical output     
+
+      ## Some analytical output
       ## Averaged G
       ivx <- round (mean (sG), 1)
       if (sieve == 'ind')
@@ -464,32 +459,32 @@ isopam <-
         ## Averaged G (only indicators)
         ivi <- round (mean (sG [xind]), 1)
         ## Number of indicators >= Gs
-        noi <- length (sG [xind])      
+        noi <- length (sG [xind])
       }
       if (sieve == TRUE)
       {
         ## Averaged G (only indicators)
         ivi <- round (mean (sG [sG >= Gs]), 1)
         ## Number of indicators >= Gs
-        noi <- length (sG [sG >= Gs])      
+        noi <- length (sG [sG >= Gs])
       }
       if (sieve == FALSE)
       {
         ## Averaged G (only indicators)
         ivi <- 'NA'
         ## Number of indicators >= Gs
-        noi <- 'NA'      
+        noi <- 'NA'
       }
 
-      ## Was this a good partition? 
+      ## Was this a good partition?
       ## At least stopcrit [1] descriptors with g >= stopcrit [2]
-      if (length (sG [sG >= stopat [2]]) >= stopat [1] * mc) fine <- TRUE                        
-      else fine <- FALSE      
-      
-      ##########################################################################      
-      
+      if (length (sG [sG >= stopat [2]]) >= stopat [1] * mc) fine <- TRUE
+      else fine <- FALSE
+
+      ##########################################################################
+
       out <- list (
-        medoids = MDS, 
+        medoids = MDS,
         clusters = CLS,
         sizes = CLI,
         is.ok = fine,
@@ -501,8 +496,8 @@ isopam <-
         ivx = ivx,
         ivi = ivi )
     }
-    else 
-    {    
+    else
+    {
 
       out <- list (
         medoids = NULL,
@@ -518,8 +513,8 @@ isopam <-
         ivi = NULL )
     }
     return (out)
-  }  
-  
+  }
+
   ## ----------- dendrogram function (code: J. Collison, 2009) -------------- ##
 
   create_dendro <- function (clust)
@@ -527,9 +522,9 @@ isopam <-
     ## Expects a list of vectors containing cluster affiliations
     ## without info about hierarchy (running number).
     ## Returns an object of class 'hclust'
-  
+
     num_obs <- length(clust[[1]])
-  
+
     dendro <- list(
       merge=array(NA, dim=c(num_obs-1,2)),
       height=rep(0, times=num_obs-1),
@@ -537,32 +532,32 @@ isopam <-
       labels=names(clust[[1]]),
       method=NULL,
       call=NULL);
-  
+
     class(dendro) <- 'hclust'
-  
-  
+
+
     ## hclust requires a set of merge operations, one observation at a time.
     ## Subsequent operations [may] refer to the index of a previous merge,
     ## so we store the operation index after everything we do.
-  
+
     opnum <- 0
-  
+
     group_opnums <- c();
-  
+
     for (level in length(clust):1)
     {
       groups <- clust[[level]];
-  
+
       groupnum <- 0
-  
+
       curlevel_group_opnums <- c();
-  
+
       repeat
       {
         groupnum <- groupnum+1
-  
+
         log_in_group <- (groups == groupnum)
-  
+
 		num_in_clust <- sum(log_in_group);
 
 		if (num_in_clust < 1)
@@ -582,9 +577,9 @@ isopam <-
           {
             if (!log_in_group[j])
               next
-  
+
             dendro$order <- c(dendro$order, j)
-  
+
             if (prev_opnum)
             {
               opnum <- opnum+1
@@ -604,7 +599,7 @@ isopam <-
               prev_index <- j
             }
           }
-  
+
           ## Done merging this cluster
           ## Save the opnum for the last merge in this cluster
           curlevel_group_opnums <- c(curlevel_group_opnums,0)
@@ -619,26 +614,26 @@ isopam <-
         else
         {
           ## higher levels
-  
+
           ## For all members of this group, see what their groupid was
           ## one level deeper and store those (uniquely)
-  
+
           groups_to_join <- c()
-  
+
           for (j in 1:num_obs)
           {
             if (!log_in_group[j])
               next
-  
+
             subgroup = clust[[level+1]][j]
-  
+
             if (sum(groups_to_join == subgroup) == 0)
             {
               ## Not there, add it
               groups_to_join <- c(groups_to_join,subgroup)
             }
           }
-  
+
           if (length(groups_to_join)>=2)
           {
             ## Merge them.
@@ -646,12 +641,12 @@ isopam <-
             for (r in 2:length(groups_to_join))
             {
               opnum <- opnum+1
-              dendro$merge[opnum,] <- c(prev_opnum, 
+              dendro$merge[opnum,] <- c(prev_opnum,
                 group_opnums[groups_to_join[r]])
               dendro$height[opnum] <- length(clust) - level + 1
               prev_opnum <- opnum;
             }
-  
+
             curlevel_group_opnums <- c(curlevel_group_opnums,0)
             curlevel_group_opnums[groupnum] <- prev_opnum
           }
@@ -665,18 +660,18 @@ isopam <-
           }
         }
       } ## end groupnum loop
-  
+
       ## Done processing this level
       group_opnums <- curlevel_group_opnums
     }
-  
+
     ## All levels have been processed.
     ## Need to merge whatever remains
-  
+
     if (length(group_opnums)>1)
     {
       prev_opnum = group_opnums[1];
-  
+
       for (r in 2:length(group_opnums))
       {
         opnum <- opnum+1
@@ -685,7 +680,7 @@ isopam <-
         prev_opnum <- opnum;
       }
     }
-  
+
     return(dendro)
   }
 
@@ -693,15 +688,15 @@ isopam <-
 
   print ('Processing level 1', quote = FALSE)
 
-  ## Prepare first partition  
+  ## Prepare first partition
   IO <- dat
   IO [IO > 0] <- 1
-  dat1 <- dat [,colSums (IO) > 1]  ## Omit species with < 2 (!) occurrences    
-  if (is.null (dim (dat1))) stop ('Not enough occurrences')      
+  dat1 <- dat [,colSums (IO) > 1]  ## Omit species with < 2 (!) occurrences
+  if (is.null (dim (dat1))) stop ('Not enough occurrences')
   dat1 <- dat1 [,apply (dat1, 2, var) > 0] ## Omit species without variance
-  if (is.null (dim (dat1))) stop ('Not enough variance in species')      
+  if (is.null (dim (dat1))) stop ('Not enough variance in species')
   dat1 <- dat1 [apply (dat1, 1, var) > 0,] ## Omit plots without variance
-  if (is.null (dim (dat1))) stop ('Not enough variance in plots')      
+  if (is.null (dim (dat1))) stop ('Not enough variance in plots')
 
   ## Initiate container for results
   matr <- matrix (NA, nrow = nrow (dat1), ncol = 1) ## Initiate cluster output
@@ -709,21 +704,21 @@ isopam <-
   colnames (matr) <- 'lev.1'
 
   ## Initiate matrix for summary ('analytics')
-  summ <- matrix (NA, nrow = 9, ncol = 1) 
-  rownames (summ) <- c('Name', 
-    'Subgroups', 
-    'Isomap.dim', 
+  summ <- matrix (NA, nrow = 9, ncol = 1)
+  rownames (summ) <- c('Name',
+    'Subgroups',
+    'Isomap.dim',
     'Isomap.k.min',
-    'Isomap.k', 
+    'Isomap.k',
     'Isomap.k.max',
     'Ind.N',
     'Ind.Gs',
-    'Global.Gs') 
+    'Global.Gs')
   colnames (summ) <- 'Part.1'
-                           
+
   ## Run core function
   output <- core (dat1)
- 
+
   ## Fill cluster container
   matr [,1] <- output$clusters
 
@@ -747,12 +742,12 @@ isopam <-
   if (l.max == 1) stepdown <- FALSE
   else stepdown <- TRUE
   ## Which group is large enough?
-  spl <- as.numeric (output$sizes > 2) 
+  spl <- as.numeric (output$sizes > 2)
   ## stop if there is nothing to split
   if (sum (spl) == 0) stepdown <- FALSE
-                 
-  ## Preparative stuff   
-  ctb <- matr ## Cluster table 
+
+  ## Preparative stuff
+  ctb <- matr ## Cluster table
   colnames (ctb) <- 'lev.1'
   mtb <- matrix (NA, nrow = nrow (dat), ncol = 0) ## Medoid table
   rownames (mtb) <- rownames (dat)
@@ -763,40 +758,40 @@ isopam <-
   names (ctb.flat) <- rownames (ctb)
 
   count <- 2 ## Counter for cluster levels
-  
+
   ## -------------- Follow-up runs ------------------------------------------ ##
-  
-  while (stepdown == TRUE) 
+
+  while (stepdown == TRUE)
   {
     if (l.max != FALSE & count > l.max) stepdown <- FALSE
     if (stepdown == TRUE)
     {
-    
+
       ifelse (sum (spl) == 1, cas <- 'group', cas <- 'groups')
-      if (cas != 0) print (paste ('Level ', count, ': Processing ', sum (spl), 
+      if (cas != 0) print (paste ('Level ', count, ': Processing ', sum (spl),
         ' ', cas, sep = ''), quote = FALSE)
       output.sub <- list ()              ## Empty list for results
       count.2 <- 1
       for (j in 1:length (spl))          ## Loop through partitions
-      {                       
-        if (spl [j] == 1) ## splittable? 
-        {                                            
+      {
+        if (spl [j] == 1) ## splittable?
+        {
           x.sub <- dat [matr [,ncol (matr)] == j,] ## Create data subset
-                                                       
+
           ## Prepare second run
           IO.x.sub <- x.sub
           IO.x.sub [IO.x.sub > 0] <- 1
-          x.sub <- x.sub [,colSums (IO.x.sub) > 1] ## Retain occurring species          
-          
-          if (is.null (dim (x.sub))) output.sub [[j]] <- NA  ## Nothing to split           
+          x.sub <- x.sub [,colSums (IO.x.sub) > 1] ## Retain occurring species
+
+          if (is.null (dim (x.sub))) output.sub [[j]] <- NA  ## Nothing to split
           else
           {
             ## Omit species and plots without variance
             x.sub <- x.sub [,apply (x.sub, 2, var) > 0]
-            x.sub <- x.sub [apply (x.sub, 1, var) > 0,]                        
-            
+            x.sub <- x.sub [apply (x.sub, 1, var) > 0,]
+
             if (nrow (x.sub) > 2) ## Enough plots left?
-            { 
+            {
               if (sum (spl) > 1) print (paste ('Group', count.2), quote = FALSE)
               output.sub [[j]] <- core (x.sub)  ## Clustering
               count.2 <- count.2 + 1
@@ -806,59 +801,59 @@ isopam <-
         }
         else output.sub [[j]] <- NA
       }
-  
+
       ## Look if some of the partitions should be rejected
       ok.vec <- vector ()
-      for (l in 1:length (spl)) 
+      for (l in 1:length (spl))
       {
-        if (is.na (output.sub [[l]][1]) == TRUE) ok.vec [[l]] <- FALSE 
+        if (is.na (output.sub [[l]][1]) == TRUE) ok.vec [[l]] <- FALSE
         else ok.vec [[l]] <- output.sub [[l]]$is.ok
       }
-  
+
       ## Report
       ok.n <- sum (as.numeric (ok.vec))
-      
+
       ## Stop if all partitions are rejected
       if (ok.n == 0) stepdown <- FALSE
-          
+
       if (stepdown == TRUE)
       {
-        ## Write new clusters to container 
+        ## Write new clusters to container
         matr <- cbind (matr, matrix (NA, nrow = nrow (matr), ncol = 1))
         colnames (matr) [ncol (matr)] <- paste ('lev.', ncol (matr), sep = '')
-        
-        for (m in 1:length (spl))  ## Fill in values 
-        {                     
-          if (ok.vec [m] == TRUE) 
+
+        for (m in 1:length (spl))  ## Fill in values
+        {
+          if (ok.vec [m] == TRUE)
           {
             sc <- output.sub [[m]]$clusters
-            idx.sc <- rownames (matr) %in% names (sc) 
+            idx.sc <- rownames (matr) %in% names (sc)
             matr [idx.sc, ncol (matr)] <- sc
           }
-        }            
-    
+        }
+
         ## Make matrix with names expressing hierarchy
-        ctb <- cbind (ctb, matrix (0, nrow = nrow (dat), ncol = 1))       
+        ctb <- cbind (ctb, matrix (0, nrow = nrow (dat), ncol = 1))
         colnames (ctb) [ncol (ctb)] <- paste ('lev.', ncol (ctb), sep = '')
         ctb.red <- na.omit (ctb)
         blub <- matr [,ncol (matr)]
         blub [is.na (blub)] <- 0
         ctb.red [,ncol (ctb.red)] <- paste (ctb.red [,ncol (ctb.red)-1],
-          blub, sep = '.')  
+          blub, sep = '.')
         ctb [rownames (ctb) %in% rownames (ctb.red), ncol (ctb)] <-
-          ctb.red [,ncol (ctb.red)]    
+          ctb.red [,ncol (ctb.red)]
         ctb <- as.data.frame (ctb)
-        
+
         ## Overwrite ctb.flat in the multiple level case
         ## Cluster affiliations without info about hierarchy (running number)
         ctb.flat <- list ()
-        for (o in 1:ncol (ctb)) 
+        for (o in 1:ncol (ctb))
         {
           ctb.flat [[o]] <- as.numeric (as.factor (ctb [,o]))
           names (ctb.flat [[o]]) <- rownames (ctb)
           names (ctb.flat) [o] <- paste ('level.', o, sep='')
         }
-    
+
         ## Make working matrix
         matr.red <- na.omit (matr)
         matr.red  [,ncol (matr.red)] <- as.numeric (as.factor ( paste
@@ -866,13 +861,13 @@ isopam <-
           sep = '.')))
         matr [rownames (matr) %in% rownames (matr.red),] <- matr.red
         matr [is.na (matr)] <- 0
-    
+
         ## Medoids
         mtb <- cbind (mtb, matrix (0, nrow = nrow (dat), ncol = 1))
         rownames (mtb) <- rownames (dat)
-        for (n in 1:length (spl)) 
+        for (n in 1:length (spl))
         {
-            if (ok.vec [n] == TRUE) 
+            if (ok.vec [n] == TRUE)
             {
               mtb [output.sub [[n]]$medoids, ncol (mtb)] <- 1
             }
@@ -881,40 +876,40 @@ isopam <-
         names (med [[count]]) <- ctb [mtb [,ncol (mtb)] == 1, ncol (ctb)]
         nam <- sort (names (med [[count]]))
         med [[count]] <- med [[count]][nam]
-        
-        ## Add new partitions to the summary matrix            
+
+        ## Add new partitions to the summary matrix
         for (x in 1:length (spl))
         {
-          if (ok.vec [x] == TRUE)  
+          if (ok.vec [x] == TRUE)
           {
             summ <- cbind (summ, matrix (NA, nrow = 9, ncol = 1))
-            colnames (summ) [ncol(summ)] <- paste ('Part.', ncol(summ), sep = '')                                      
+            colnames (summ) [ncol(summ)] <- paste ('Part.', ncol(summ), sep = '')
             summ [2,ncol(summ)] <- length (output.sub[[x]]$medoids) ## Subgroups
             summ [3,ncol(summ)] <- output.sub[[x]]$d             ## Isomap dims
-            summ [4,ncol(summ)] <- output.sub[[x]]$k.min         ## Minimum k            
-            summ [5,ncol(summ)] <- output.sub[[x]]$k             ## Selected k             
-            summ [6,ncol(summ)] <- output.sub[[x]]$k.max         ## Maximum k            
-            summ [7,ncol(summ)] <- output.sub[[x]]$noi           ## Indicators    
-            summ [8,ncol(summ)] <- output.sub[[x]]$ivi           ## IV (Indic.)    
-            summ [9,ncol(summ)] <- output.sub[[x]]$ivx           ## IV (all)    
+            summ [4,ncol(summ)] <- output.sub[[x]]$k.min         ## Minimum k
+            summ [5,ncol(summ)] <- output.sub[[x]]$k             ## Selected k
+            summ [6,ncol(summ)] <- output.sub[[x]]$k.max         ## Maximum k
+            summ [7,ncol(summ)] <- output.sub[[x]]$noi           ## Indicators
+            summ [8,ncol(summ)] <- output.sub[[x]]$ivi           ## IV (Indic.)
+            summ [9,ncol(summ)] <- output.sub[[x]]$ivx           ## IV (all)
           }
         }
-  
+
         ## OK, now we have a matrix with cluster affiliations of this level
         ## Repeat the search for splittable units:
-    
-        subtab <- table (matr [,ncol (matr)])                    
-        subtab <- subtab [names (subtab) != '0'] 
-        spl <- as.numeric (subtab > 2)                                
+
+        subtab <- table (matr [,ncol (matr)])
+        subtab <- subtab [names (subtab) != '0']
+        spl <- as.numeric (subtab > 2)
         names (spl) <- names (subtab)
-               
+
         count <- count + 1
-        
+
         ## Create object of class 'hclust'
         dendro <- create_dendro(ctb.flat)
       }
-    }     
-  } ## End while-loop    
+    }
+  } ## End while-loop
 
   ## Cluster names
   if (ncol (ctb) > 1)
@@ -929,15 +924,15 @@ isopam <-
     for (z in 1:length (cnam))
     {
       nch <- nchar (cnam [z])
-      childs <- c (childs, length (grep (cnam [z], substr (cnam, 1, nch), 
+      childs <- c (childs, length (grep (cnam [z], substr (cnam, 1, nch),
         value = TRUE)))
     }
-    summ [1,2:ncol(summ)] <- cnam [childs > 1]  
+    summ [1,2:ncol(summ)] <- cnam [childs > 1]
   }
   summ <- as.data.frame (summ)
 
   ## ------------ Output ---------------------------------------------------- ##
-  
+
   if (ncol (ctb) == 1)
     OUT <- list (
       call = sys.call (),
@@ -962,24 +957,24 @@ isopam <-
     )
 
     class (OUT) <- 'isopam'
-    
+
   ## save output
   if (juice == TRUE)
   {
-    write.table (ctb, file = 'isopam/juicein.txt', 
-      col.names = FALSE, quote = FALSE)  
+    write.table (ctb, file = 'isopam/juicein.txt',
+      col.names = FALSE, quote = FALSE)
     if (!is.null (OUT$dendro [1]))
     {
       wth <- (nrow (ctb) * 11) + 100
       bmp (filename = 'isopam/juicetree.bmp', width = wth)
       plot (OUT$dendro)
       dev.off()
-    }  
+    }
   }
-  if (ncol (ctb) == 1) 
+  if (ncol (ctb) == 1)
     print (paste ("Non-hierarchical partition created"), quote = FALSE)
-  else 
-    print (paste ("Cluster tree with", ncol (ctb), "levels created"), 
+  else
+    print (paste ("Cluster tree with", ncol (ctb), "levels created"),
     quote = FALSE)
 
   invisible (OUT)
@@ -988,18 +983,17 @@ isopam <-
 plot.isopam <-
 function (x, ...)
 {
-  if (!is.null (x$dendro [1])) 
-  {  
+  if (!is.null (x$dendro [1]))
+  {
     tree <- x$dendro
-    plot (tree, main = format(x$call), ...)  
-  }     
+    plot (tree, main = format(x$call), ...)
+  }
   else print ('No cluster hierarchy - nothing to plot', quote = FALSE)
 }
 
 identify.isopam <-
 function (x, ...)
 {
-  identify (x$dendro, MAXCLUSTER = nrow (x$dat), ...)  
+  identify (x$dendro, MAXCLUSTER = nrow (x$dat), ...)
 }
 
-   
\ No newline at end of file
